import base64
import logging

try:
    from Crypto.Cipher import AES
    from Crypto import Random
except ImportError:
    from .nocrypto import AES, Random

from ably.sync.types.typedbuffer import TypedBuffer
from ably.sync.util.exceptions import AblyException

log = logging.getLogger(__name__)


class CipherParams:
    def __init__(self, algorithm='AES', mode='CBC', secret_key=None, iv=None):
        self.__algorithm = algorithm.upper()
        self.__secret_key = secret_key
        self.__key_length = len(secret_key) * 8 if secret_key is not None else 128
        self.__mode = mode.upper()
        self.__iv = iv

    @property
    def algorithm(self):
        return self.__algorithm

    @property
    def secret_key(self):
        return self.__secret_key

    @property
    def iv(self):
        return self.__iv

    @property
    def key_length(self):
        return self.__key_length

    @property
    def mode(self):
        return self.__mode


class CbcChannelCipher:
    def __init__(self, cipher_params):
        self.__secret_key = (cipher_params.secret_key or
                             self.__random(cipher_params.key_length / 8))
        if isinstance(self.__secret_key, str):
            self.__secret_key = self.__secret_key.encode()
        self.__iv = cipher_params.iv or self.__random(16)
        self.__block_size = len(self.__iv)
        if cipher_params.algorithm != 'AES':
            raise NotImplementedError('Only AES algorithm is supported')
        self.__algorithm = cipher_params.algorithm
        if cipher_params.mode != 'CBC':
            raise NotImplementedError('Only CBC mode is supported')
        self.__mode = cipher_params.mode
        self.__key_length = cipher_params.key_length
        self.__encryptor = AES.new(self.__secret_key, AES.MODE_CBC, self.__iv)

    def __pad(self, data):
        padding_size = self.__block_size - (len(data) % self.__block_size)

        padding_char = bytes((padding_size,))
        padded = data + padding_char * padding_size

        return padded

    def __unpad(self, data):
        padding_size = data[-1]

        if padding_size > len(data):
            # Too short
            raise AblyException('invalid-padding', 0, 0)

        if padding_size == 0:
            # Missing padding
            raise AblyException('invalid-padding', 0, 0)

        for i in range(padding_size):
            # Invalid padding bytes
            if padding_size != data[-i - 1]:
                raise AblyException('invalid-padding', 0, 0)

        return data[:-padding_size]

    def __random(self, length):
        rndfile = Random.new()
        return rndfile.read(length)

    def encrypt(self, plaintext):
        if isinstance(plaintext, bytearray):
            plaintext = bytes(plaintext)
        padded_plaintext = self.__pad(plaintext)
        encrypted = self.__iv + self.__encryptor.encrypt(padded_plaintext)
        self.__iv = encrypted[-self.__block_size:]
        return encrypted

    def decrypt(self, ciphertext):
        if isinstance(ciphertext, bytearray):
            ciphertext = bytes(ciphertext)
        iv = ciphertext[:self.__block_size]
        ciphertext = ciphertext[self.__block_size:]
        decryptor = AES.new(self.__secret_key, AES.MODE_CBC, iv)
        decrypted = decryptor.decrypt(ciphertext)
        return bytearray(self.__unpad(decrypted))

    @property
    def secret_key(self):
        return self.__secret_key

    @property
    def iv(self):
        return self.__iv

    @property
    def cipher_type(self):
        return ("%s-%s-%s" % (self.__algorithm, self.__key_length,
                self.__mode)).lower()


class CipherData(TypedBuffer):
    ENCODING_ID = 'cipher'

    def __init__(self, buffer, type, cipher_type=None, **kwargs):
        self.__cipher_type = cipher_type
        super().__init__(buffer, type, **kwargs)

    @property
    def encoding_str(self):
        return self.ENCODING_ID + '+' + self.__cipher_type


DEFAULT_KEYLENGTH = 256
DEFAULT_BLOCKLENGTH = 16


def generate_random_key(length=DEFAULT_KEYLENGTH):
    rndfile = Random.new()
    return rndfile.read(length // 8)


def get_default_params(params=None):
    if type(params) in [str, bytes]:
        raise ValueError("Calling get_default_params with a key directly is deprecated, it expects a params dict")

    key = params.get('key')
    algorithm = params.get('algorithm') or 'AES'
    iv = params.get('iv') or generate_random_key(DEFAULT_BLOCKLENGTH * 8)
    mode = params.get('mode') or 'CBC'

    if not key:
        raise ValueError("Crypto.get_default_params: a key is required")

    if type(key) == str:
        key = base64.b64decode(key)

    cipher_params = CipherParams(algorithm=algorithm, secret_key=key, iv=iv, mode=mode)
    validate_cipher_params(cipher_params)
    return cipher_params


def get_cipher(params):
    if isinstance(params, CipherParams):
        cipher_params = params
    else:
        cipher_params = get_default_params(params)
    return CbcChannelCipher(cipher_params)


def validate_cipher_params(cipher_params):
    if cipher_params.algorithm == 'AES' and cipher_params.mode == 'CBC':
        key_length = cipher_params.key_length
        if key_length == 128 or key_length == 256:
            return
        raise ValueError(
            'Unsupported key length %s for aes-cbc encryption. Encryption key must be 128 or 256 bits'
            ' (16 or 32 ASCII characters)' % key_length)



import functools
import logging


log = logging.getLogger(__name__)


class AblyException(Exception):
    def __new__(cls, message, status_code, code, cause=None):
        if cls == AblyException and status_code == 401:
            return AblyAuthException(message, status_code, code, cause)
        return super().__new__(cls, message, status_code, code, cause)

    def __init__(self, message, status_code, code, cause=None):
        super().__init__()
        self.message = message
        self.code = code
        self.status_code = status_code
        self.cause = cause

    def __str__(self):
        str = '%s %s %s' % (self.code, self.status_code, self.message)
        if self.cause is not None:
            str += ' (cause: %s)' % self.cause
        return str

    @property
    def is_server_error(self):
        return 500 <= self.status_code <= 599

    @staticmethod
    def raise_for_response(response):
        if 200 <= response.status_code < 300:
            # Valid response
            return

        try:
            json_response = response.json()
        except Exception:
            log.debug("Response not json: %d %s",
                      response.status_code,
                      response.text)
            raise AblyException(message=response.text,
                                status_code=response.status_code,
                                code=response.status_code * 100)

        if json_response and 'error' in json_response:
            error = json_response['error']
            try:
                raise AblyException(
                    message=error['message'],
                    status_code=error['statusCode'],
                    code=int(error['code']),
                )
            except KeyError:
                msg = "Unexpected exception decoding server response: %s"
                msg = msg % response.text
                raise AblyException(message=msg, status_code=500, code=50000)

        raise AblyException(message="",
                            status_code=response.status_code,
                            code=response.status_code * 100)

    @staticmethod
    def from_exception(e):
        if isinstance(e, AblyException):
            return e
        return AblyException("Unexpected exception: %s" % e, 500, 50000)

    @staticmethod
    def from_dict(value: dict):
        return AblyException(value.get('message'), value.get('statusCode'), value.get('code'))


def catch_all(func):
    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except Exception as e:
            log.exception(e)
            raise AblyException.from_exception(e)

    return wrapper


class AblyAuthException(AblyException):
    pass


class IncompatibleClientIdException(AblyException):
    pass



class Defaults:
    protocol_version = "2"
    fallback_hosts = [
        "a.ably-realtime.com",
        "b.ably-realtime.com",
        "c.ably-realtime.com",
        "d.ably-realtime.com",
        "e.ably-realtime.com",
    ]

    rest_host = "rest.ably.io"
    realtime_host = "realtime.ably.io"  # RTN2
    connectivity_check_url = "https://internet-up.ably-realtime.com/is-the-internet-up.txt"
    environment = 'production'

    port = 80
    tls_port = 443
    connect_timeout = 15000
    disconnect_timeout = 10000
    suspended_timeout = 60000
    comet_recv_timeout = 90000
    comet_send_timeout = 10000
    realtime_request_timeout = 10000
    channel_retry_timeout = 15000
    disconnected_retry_timeout = 15000
    connection_state_ttl = 120000
    suspended_retry_timeout = 30000

    transports = []  # ["web_socket", "comet"]

    http_max_retry_count = 3

    fallback_retry_timeout = 600000  # 10min

    @staticmethod
    def get_port(options):
        if options.tls:
            if options.tls_port:
                return options.tls_port
            else:
                return Defaults.tls_port
        else:
            if options.port:
                return options.port
            else:
                return Defaults.port

    @staticmethod
    def get_scheme(options):
        if options.tls:
            return "https"
        else:
            return "http"

    @staticmethod
    def get_environment_fallback_hosts(environment):
        return [
            environment + "-a-fallback.ably-realtime.com",
            environment + "-b-fallback.ably-realtime.com",
            environment + "-c-fallback.ably-realtime.com",
            environment + "-d-fallback.ably-realtime.com",
            environment + "-e-fallback.ably-realtime.com",
        ]



import json
import time

from ably.types.capability import Capability


class TokenDetails:

    DEFAULTS = {'ttl': 60 * 60 * 1000}
    # Buffer in milliseconds before a token is considered unusable
    # For example, if buffer is 10000ms, the token can no longer be used for
    # new requests 9000ms before it expires
    TOKEN_EXPIRY_BUFFER = 15 * 1000

    def __init__(self, token=None, expires=None, issued=0,
                 capability=None, client_id=None):
        if expires is None:
            self.__expires = time.time() * 1000 + TokenDetails.DEFAULTS['ttl']
        else:
            self.__expires = expires
        self.__token = token
        self.__issued = issued
        if capability and isinstance(capability, str):
            try:
                self.__capability = Capability(json.loads(capability))
            except json.JSONDecodeError:
                self.__capability = Capability(json.loads(capability.replace("'", '"')))
        else:
            self.__capability = Capability(capability or {})
        self.__client_id = client_id

    @property
    def token(self):
        return self.__token

    @property
    def expires(self):
        return self.__expires

    @property
    def issued(self):
        return self.__issued

    @property
    def capability(self):
        return self.__capability

    @property
    def client_id(self):
        return self.__client_id

    def to_dict(self):
        return {
            'expires': self.expires,
            'token': self.token,
            'issued': self.issued,
            'capability': self.capability.to_dict(),
            'clientId': self.client_id,
        }

    @staticmethod
    def from_dict(obj):
        kwargs = {
            'token': obj.get("token"),
            'capability': obj.get("capability"),
            'client_id': obj.get("clientId")
        }
        expires = obj.get("expires")
        kwargs['expires'] = expires if expires is None else int(expires)
        issued = obj.get("issued")
        kwargs['issued'] = issued if issued is None else int(issued)

        return TokenDetails(**kwargs)

    @staticmethod
    def from_json(data):
        if isinstance(data, str):
            data = json.loads(data)

        mapping = {
            'clientId': 'client_id',
        }
        for name in data:
            py_name = mapping.get(name)
            if py_name:
                data[py_name] = data.pop(name)

        return TokenDetails(**data)

    def __eq__(self, other):
        if isinstance(other, TokenDetails):
            return (self.expires == other.expires
                    and self.token == other.token
                    and self.issued == other.issued
                    and self.capability == other.capability
                    and self.client_id == other.client_id)
        return NotImplemented



from enum import Enum
from dataclasses import dataclass
from typing import Optional

from ably.util.exceptions import AblyException


class ConnectionState(str, Enum):
    INITIALIZED = 'initialized'
    CONNECTING = 'connecting'
    CONNECTED = 'connected'
    DISCONNECTED = 'disconnected'
    CLOSING = 'closing'
    CLOSED = 'closed'
    FAILED = 'failed'
    SUSPENDED = 'suspended'


class ConnectionEvent(str, Enum):
    INITIALIZED = 'initialized'
    CONNECTING = 'connecting'
    CONNECTED = 'connected'
    DISCONNECTED = 'disconnected'
    CLOSING = 'closing'
    CLOSED = 'closed'
    FAILED = 'failed'
    SUSPENDED = 'suspended'
    UPDATE = 'update'


@dataclass
class ConnectionStateChange:
    previous: ConnectionState
    current: ConnectionState
    event: ConnectionEvent
    reason: Optional[AblyException] = None  # RTN4f



import logging
from datetime import datetime

log = logging.getLogger(__name__)


class Stats:

    def __init__(self, entries=None, unit=None, interval_id=None, in_progress=None, app_id=None, schema=None):
        self.interval_id = interval_id or ''
        self.entries = entries
        self.unit = unit
        self.interval_time = interval_from_interval_id(self.interval_id)
        self.in_progress = in_progress
        self.app_id = app_id
        self.schema = schema

    @classmethod
    def from_dict(cls, stats_dict):
        stats_dict = stats_dict or {}

        kwargs = {
            "entries": stats_dict.get("entries"),
            "unit": stats_dict.get("unit"),
            "interval_id": stats_dict.get("intervalId"),
            "in_progress": stats_dict.get("inProgress"),
            "app_id": stats_dict.get("appId"),
            "schema": stats_dict.get("schema"),
        }

        return cls(**kwargs)

    @classmethod
    def from_array(cls, stats_array):
        return [cls.from_dict(d) for d in stats_array]

    @staticmethod
    def to_interval_id(date_time, granularity):
        return date_time.strftime(INTERVALS_FMT[granularity])


def stats_response_processor(response):
    stats_array = response.to_native()
    return Stats.from_array(stats_array)


INTERVALS_FMT = {
    'minute': '%Y-%m-%d:%H:%M',
    'hour': '%Y-%m-%d:%H',
    'day': '%Y-%m-%d',
    'month': '%Y-%m',
}


def granularity_from_interval_id(interval_id):
    for key, value in INTERVALS_FMT.items():
        try:
            datetime.strptime(interval_id, value)
            return key
        except ValueError:
            pass
    raise ValueError("Unsupported intervalId")


def interval_from_interval_id(interval_id):
    granularity = granularity_from_interval_id(interval_id)
    return datetime.strptime(interval_id, INTERVALS_FMT[granularity])



# This functionality is depreceated and will be removed
# Message Pack is the replacement for all binary data messages

import json
import struct


class DataType:
    NONE = 0
    TRUE = 1
    FALSE = 2
    INT32 = 3
    INT64 = 4
    DOUBLE = 5
    STRING = 6
    BUFFER = 7
    JSONARRAY = 8
    JSONOBJECT = 9


class Limits:
    INT32_MAX = 2 ** 31
    INT32_MIN = -(2 ** 31 + 1)
    INT64_MAX = 2 ** 63
    INT64_MIN = - (2 ** 63 + 1)


_decoders = {DataType.TRUE: lambda b: True,
             DataType.FALSE: lambda b: False,
             DataType.INT32: lambda b: struct.unpack('>i', b)[0],
             DataType.INT64: lambda b: struct.unpack('>q', b)[0],
             DataType.DOUBLE: lambda b: struct.unpack('>d', b)[0],
             DataType.STRING: lambda b: b.decode('utf-8'),
             DataType.BUFFER: lambda b: b,
             DataType.JSONARRAY: lambda b: json.loads(b.decode('utf-8')),
             DataType.JSONOBJECT: lambda b: json.loads(b.decode('utf-8'))}


class TypedBuffer:
    def __init__(self, buffer, type):
        self.__buffer = buffer
        self.__type = type

    def __eq__(self, other):
        if isinstance(other, TypedBuffer):
            return self.buffer == other.buffer and self.type == other.type
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, TypedBuffer):
            result = self.__eq__(other)
            if result != NotImplemented:
                return not result
        return NotImplemented

    @staticmethod
    def from_obj(obj):
        if isinstance(obj, TypedBuffer):
            return obj
        elif isinstance(obj, (bytes, bytearray)):
            data_type = DataType.BUFFER
            buffer = obj
        elif isinstance(obj, str):
            data_type = DataType.STRING
            buffer = obj.encode('utf-8')
        elif isinstance(obj, bool):
            data_type = DataType.TRUE if obj else DataType.FALSE
            buffer = None
        elif isinstance(obj, int):
            if Limits.INT32_MIN <= obj <= Limits.INT32_MAX:
                data_type = DataType.INT32
                buffer = struct.pack('>i', obj)
            elif Limits.INT64_MIN <= obj <= Limits.INT64_MAX:
                data_type = DataType.INT64
                buffer = struct.pack('>q', obj)
            else:
                raise ValueError('Number too large %d' % obj)
        elif isinstance(obj, float):
            data_type = DataType.DOUBLE
            buffer = struct.pack('>d', obj)
        elif isinstance(obj, list):
            data_type = DataType.JSONARRAY
            buffer = json.dumps(obj, separators=(',', ':')).encode('utf-8')
        elif isinstance(obj, dict):
            data_type = DataType.JSONOBJECT
            buffer = json.dumps(obj, separators=(',', ':')).encode('utf-8')
        else:
            raise TypeError('Unexpected object type %s' % type(obj))

        return TypedBuffer(buffer, data_type)

    @property
    def buffer(self):
        return self.__buffer

    @property
    def type(self):
        return self.__type

    def decode(self):
        decoder = _decoders.get(self.type)
        if decoder is not None:
            return decoder(self.buffer)
        raise ValueError('Unsupported data type %s' % self.type)



from __future__ import annotations
import asyncio
import logging
from typing import Optional, TYPE_CHECKING
from ably.realtime.connection import ConnectionState
from ably.transport.websockettransport import ProtocolMessageAction
from ably.rest.channel import Channel, Channels as RestChannels
from ably.types.channelstate import ChannelState, ChannelStateChange
from ably.types.flags import Flag, has_flag
from ably.types.message import Message
from ably.util.eventemitter import EventEmitter
from ably.util.exceptions import AblyException
from ably.util.helper import Timer, is_callable_or_coroutine

if TYPE_CHECKING:
    from ably.realtime.realtime import AblyRealtime

log = logging.getLogger(__name__)


class RealtimeChannel(EventEmitter, Channel):
    """
    Ably Realtime Channel

    Attributes
    ----------
    name: str
        Channel name
    state: str
        Channel state
    error_reason: AblyException
        An AblyException instance describing the last error which occurred on the channel, if any.

    Methods
    -------
    attach()
        Attach to channel
    detach()
        Detach from channel
    subscribe(*args)
        Subscribe to messages on a channel
    unsubscribe(*args)
        Unsubscribe to messages from a channel
    """

    def __init__(self, realtime: AblyRealtime, name: str):
        EventEmitter.__init__(self)
        self.__name = name
        self.__realtime = realtime
        self.__state = ChannelState.INITIALIZED
        self.__message_emitter = EventEmitter()
        self.__state_timer: Optional[Timer] = None
        self.__attach_resume = False
        self.__channel_serial: Optional[str] = None
        self.__retry_timer: Optional[Timer] = None
        self.__error_reason: Optional[AblyException] = None

        # Used to listen to state changes internally, if we use the public event emitter interface then internals
        # will be disrupted if the user called .off() to remove all listeners
        self.__internal_state_emitter = EventEmitter()

        Channel.__init__(self, realtime, name, {})

    # RTL4
    async def attach(self) -> None:
        """Attach to channel

        Attach to this channel ensuring the channel is created in the Ably system and all messages published
        on the channel are received by any channel listeners registered using subscribe

        Raises
        ------
        AblyException
            If unable to attach channel
        """

        log.info(f'RealtimeChannel.attach() called, channel = {self.name}')

        # RTL4a - if channel is attached do nothing
        if self.state == ChannelState.ATTACHED:
            return

        self.__error_reason = None

        # RTL4b
        if self.__realtime.connection.state not in [
            ConnectionState.CONNECTING,
            ConnectionState.CONNECTED,
            ConnectionState.DISCONNECTED
        ]:
            raise AblyException(
                message=f"Unable to attach; channel state = {self.state}",
                code=90001,
                status_code=400
            )

        if self.state != ChannelState.ATTACHING:
            self._request_state(ChannelState.ATTACHING)

        state_change = await self.__internal_state_emitter.once_async()

        if state_change.current in (ChannelState.SUSPENDED, ChannelState.FAILED):
            raise state_change.reason

    def _attach_impl(self):
        log.debug("RealtimeChannel.attach_impl(): sending ATTACH protocol message")

        # RTL4c
        attach_msg = {
            "action": ProtocolMessageAction.ATTACH,
            "channel": self.name,
        }

        if self.__attach_resume:
            attach_msg["flags"] = Flag.ATTACH_RESUME
        if self.__channel_serial:
            attach_msg["channelSerial"] = self.__channel_serial

        self._send_message(attach_msg)

    # RTL5
    async def detach(self) -> None:
        """Detach from channel

        Any resulting channel state change is emitted to any listeners registered
        Once all clients globally have detached from the channel, the channel will be released
        in the Ably service within two minutes.

        Raises
        ------
        AblyException
            If unable to detach channel
        """

        log.info(f'RealtimeChannel.detach() called, channel = {self.name}')

        # RTL5g, RTL5b - raise exception if state invalid
        if self.__realtime.connection.state in [ConnectionState.CLOSING, ConnectionState.FAILED]:
            raise AblyException(
                message=f"Unable to detach; channel state = {self.state}",
                code=90001,
                status_code=400
            )

        # RTL5a - if channel already detached do nothing
        if self.state in [ChannelState.INITIALIZED, ChannelState.DETACHED]:
            return

        if self.state == ChannelState.SUSPENDED:
            self._notify_state(ChannelState.DETACHED)
            return
        elif self.state == ChannelState.FAILED:
            raise AblyException("Unable to detach; channel state = failed", 90001, 400)
        else:
            self._request_state(ChannelState.DETACHING)

        # RTL5h - wait for pending connection
        if self.__realtime.connection.state == ConnectionState.CONNECTING:
            await self.__realtime.connect()

        state_change = await self.__internal_state_emitter.once_async()
        new_state = state_change.current

        if new_state == ChannelState.DETACHED:
            return
        elif new_state == ChannelState.ATTACHING:
            raise AblyException("Detach request superseded by a subsequent attach request", 90000, 409)
        else:
            raise state_change.reason

    def _detach_impl(self) -> None:
        log.debug("RealtimeChannel.detach_impl(): sending DETACH protocol message")

        # RTL5d
        detach_msg = {
            "action": ProtocolMessageAction.DETACH,
            "channel": self.__name,
        }

        self._send_message(detach_msg)

    # RTL7
    async def subscribe(self, *args) -> None:
        """Subscribe to a channel

        Registers a listener for messages on the channel.
        The caller supplies a listener function, which is called
        each time one or more messages arrives on the channel.

        The function resolves once the channel is attached.

        Parameters
        ----------
        *args: event, listener
            Subscribe event and listener

            arg1(event): str, optional
                Subscribe to messages with the given event name

            arg2(listener): callable
                Subscribe to all messages on the channel

            When no event is provided, arg1 is used as the listener.

        Raises
        ------
        AblyException
            If unable to subscribe to a channel due to invalid connection state
        ValueError
            If no valid subscribe arguments are passed
        """
        if isinstance(args[0], str):
            event = args[0]
            if not args[1]:
                raise ValueError("channel.subscribe called without listener")
            if not is_callable_or_coroutine(args[1]):
                raise ValueError("subscribe listener must be function or coroutine function")
            listener = args[1]
        elif is_callable_or_coroutine(args[0]):
            listener = args[0]
            event = None
        else:
            raise ValueError('invalid subscribe arguments')

        log.info(f'RealtimeChannel.subscribe called, channel = {self.name}, event = {event}')

        if event is not None:
            # RTL7b
            self.__message_emitter.on(event, listener)
        else:
            # RTL7a
            self.__message_emitter.on(listener)

        # RTL7c
        await self.attach()

    # RTL8
    def unsubscribe(self, *args) -> None:
        """Unsubscribe from a channel

        Deregister the given listener for (for any/all event names).
        This removes an earlier event-specific subscription.

        Parameters
        ----------
        *args: event, listener
            Unsubscribe event and listener

            arg1(event): str, optional
                Unsubscribe to messages with the given event name

            arg2(listener): callable
                Unsubscribe to all messages on the channel

            When no event is provided, arg1 is used as the listener.

        Raises
        ------
        ValueError
            If no valid unsubscribe arguments are passed, no listener or listener is not a function
            or coroutine
        """
        if len(args) == 0:
            event = None
            listener = None
        elif isinstance(args[0], str):
            event = args[0]
            if not args[1]:
                raise ValueError("channel.unsubscribe called without listener")
            if not is_callable_or_coroutine(args[1]):
                raise ValueError("unsubscribe listener must be a function or coroutine function")
            listener = args[1]
        elif is_callable_or_coroutine(args[0]):
            listener = args[0]
            event = None
        else:
            raise ValueError('invalid unsubscribe arguments')

        log.info(f'RealtimeChannel.unsubscribe called, channel = {self.name}, event = {event}')

        if listener is None:
            # RTL8c
            self.__message_emitter.off()
        elif event is not None:
            # RTL8b
            self.__message_emitter.off(event, listener)
        else:
            # RTL8a
            self.__message_emitter.off(listener)

    def _on_message(self, proto_msg: dict) -> None:
        action = proto_msg.get('action')
        # RTL4c1
        channel_serial = proto_msg.get('channelSerial')
        if channel_serial:
            self.__channel_serial = channel_serial
        # TM2a, TM2c, TM2f
        Message.update_inner_message_fields(proto_msg)

        if action == ProtocolMessageAction.ATTACHED:
            flags = proto_msg.get('flags')
            error = proto_msg.get("error")
            exception = None
            resumed = False

            if error:
                exception = AblyException.from_dict(error)

            if flags:
                resumed = has_flag(flags, Flag.RESUMED)

            #  RTL12
            if self.state == ChannelState.ATTACHED:
                if not resumed:
                    state_change = ChannelStateChange(self.state, ChannelState.ATTACHED, resumed, exception)
                    self._emit("update", state_change)
            elif self.state == ChannelState.ATTACHING:
                self._notify_state(ChannelState.ATTACHED, resumed=resumed)
            else:
                log.warn("RealtimeChannel._on_message(): ATTACHED received while not attaching")
        elif action == ProtocolMessageAction.DETACHED:
            if self.state == ChannelState.DETACHING:
                self._notify_state(ChannelState.DETACHED)
            elif self.state == ChannelState.ATTACHING:
                self._notify_state(ChannelState.SUSPENDED)
            else:
                self._request_state(ChannelState.ATTACHING)
        elif action == ProtocolMessageAction.MESSAGE:
            messages = Message.from_encoded_array(proto_msg.get('messages'))
            for message in messages:
                self.__message_emitter._emit(message.name, message)
        elif action == ProtocolMessageAction.ERROR:
            error = AblyException.from_dict(proto_msg.get('error'))
            self._notify_state(ChannelState.FAILED, reason=error)

    def _request_state(self, state: ChannelState) -> None:
        log.debug(f'RealtimeChannel._request_state(): state = {state}')
        self._notify_state(state)
        self._check_pending_state()

    def _notify_state(self, state: ChannelState, reason: Optional[AblyException] = None,
                      resumed: bool = False) -> None:
        log.debug(f'RealtimeChannel._notify_state(): state = {state}')

        self.__clear_state_timer()

        if state == self.state:
            return

        if reason is not None:
            self.__error_reason = reason

        if state == ChannelState.INITIALIZED:
            self.__error_reason = None

        if state == ChannelState.SUSPENDED and self.ably.connection.state == ConnectionState.CONNECTED:
            self.__start_retry_timer()
        else:
            self.__cancel_retry_timer()

        # RTL4j1
        if state == ChannelState.ATTACHED:
            self.__attach_resume = True
        if state in (ChannelState.DETACHING, ChannelState.FAILED):
            self.__attach_resume = False

        # RTP5a1
        if state in (ChannelState.DETACHED, ChannelState.SUSPENDED, ChannelState.FAILED):
            self.__channel_serial = None

        state_change = ChannelStateChange(self.__state, state, resumed, reason=reason)

        self.__state = state
        self._emit(state, state_change)
        self.__internal_state_emitter._emit(state, state_change)

    def _send_message(self, msg: dict) -> None:
        asyncio.create_task(self.__realtime.connection.connection_manager.send_protocol_message(msg))

    def _check_pending_state(self):
        connection_state = self.__realtime.connection.connection_manager.state

        if connection_state is not ConnectionState.CONNECTED:
            log.debug(f"RealtimeChannel._check_pending_state(): connection state = {connection_state}")
            return

        if self.state == ChannelState.ATTACHING:
            self.__start_state_timer()
            self._attach_impl()
        elif self.state == ChannelState.DETACHING:
            self.__start_state_timer()
            self._detach_impl()

    def __start_state_timer(self) -> None:
        if not self.__state_timer:
            def on_timeout() -> None:
                log.debug('RealtimeChannel.start_state_timer(): timer expired')
                self.__state_timer = None
                self.__timeout_pending_state()

            self.__state_timer = Timer(self.__realtime.options.realtime_request_timeout, on_timeout)

    def __clear_state_timer(self) -> None:
        if self.__state_timer:
            self.__state_timer.cancel()
            self.__state_timer = None

    def __timeout_pending_state(self) -> None:
        if self.state == ChannelState.ATTACHING:
            self._notify_state(
                ChannelState.SUSPENDED, reason=AblyException("Channel attach timed out", 408, 90007))
        elif self.state == ChannelState.DETACHING:
            self._notify_state(ChannelState.ATTACHED, reason=AblyException("Channel detach timed out", 408, 90007))
        else:
            self._check_pending_state()

    def __start_retry_timer(self) -> None:
        if self.__retry_timer:
            return

        self.__retry_timer = Timer(self.ably.options.channel_retry_timeout, self.__on_retry_timer_expire)

    def __cancel_retry_timer(self) -> None:
        if self.__retry_timer:
            self.__retry_timer.cancel()
            self.__retry_timer = None

    def __on_retry_timer_expire(self) -> None:
        if self.state == ChannelState.SUSPENDED and self.ably.connection.state == ConnectionState.CONNECTED:
            self.__retry_timer = None
            log.info("RealtimeChannel retry timer expired, attempting a new attach")
            self._request_state(ChannelState.ATTACHING)

    # RTL23
    @property
    def name(self) -> str:
        """Returns channel name"""
        return self.__name

    # RTL2b
    @property
    def state(self) -> ChannelState:
        """Returns channel state"""
        return self.__state

    @state.setter
    def state(self, state: ChannelState) -> None:
        self.__state = state

    # RTL24
    @property
    def error_reason(self) -> Optional[AblyException]:
        """An AblyException instance describing the last error which occurred on the channel, if any."""
        return self.__error_reason


class Channels(RestChannels):
    """Creates and destroys RealtimeChannel objects.

    Methods
    -------
    get(name)
        Gets a channel
    release(name)
        Releases a channel
    """

    # RTS3
    def get(self, name: str) -> RealtimeChannel:
        """Creates a new RealtimeChannel object, or returns the existing channel object.

        Parameters
        ----------

        name: str
            Channel name
        """
        if name not in self.__all:
            channel = self.__all[name] = RealtimeChannel(self.__ably, name)
        else:
            channel = self.__all[name]
        return channel

    # RTS4
    def release(self, name: str) -> None:
        """Releases a RealtimeChannel object, deleting it, and enabling it to be garbage collected

        It also removes any listeners associated with the channel.
        To release a channel, the channel state must be INITIALIZED, DETACHED, or FAILED.


        Parameters
        ----------
        name: str
            Channel name
        """
        if name not in self.__all:
            return
        del self.__all[name]

    def _on_channel_message(self, msg: dict) -> None:
        channel_name = msg.get('channel')
        if not channel_name:
            log.error(
                'Channels.on_channel_message()',
                f'received event without channel, action = {msg.get("action")}'
            )
            return

        channel = self.__all[channel_name]
        if not channel:
            log.warning(
                'Channels.on_channel_message()',
                f'receieved event for non-existent channel: {channel_name}'
            )
            return

        channel._on_message(msg)

    def _propagate_connection_interruption(self, state: ConnectionState, reason: Optional[AblyException]) -> None:
        from_channel_states = (
            ChannelState.ATTACHING,
            ChannelState.ATTACHED,
            ChannelState.DETACHING,
            ChannelState.SUSPENDED,
        )

        connection_to_channel_state = {
            ConnectionState.CLOSING: ChannelState.DETACHED,
            ConnectionState.CLOSED: ChannelState.DETACHED,
            ConnectionState.FAILED: ChannelState.FAILED,
            ConnectionState.SUSPENDED: ChannelState.SUSPENDED,
        }

        for channel_name in self.__all:
            channel = self.__all[channel_name]
            if channel.state in from_channel_states:
                channel._notify_state(connection_to_channel_state[state], reason)

    def _on_connected(self) -> None:
        for channel_name in self.__all:
            channel = self.__all[channel_name]
            if channel.state == ChannelState.ATTACHING or channel.state == ChannelState.DETACHING:
                channel._check_pending_state()
            elif channel.state == ChannelState.SUSPENDED:
                asyncio.create_task(channel.attach())
            elif channel.state == ChannelState.ATTACHED:
                channel._request_state(ChannelState.ATTACHING)

    def _initialize_channels(self) -> None:
        for channel_name in self.__all:
            channel = self.__all[channel_name]
            channel._request_state(ChannelState.INITIALIZED)



import base64
import os
import platform

import ably


class HttpUtils:
    default_format = "json"

    mime_types = {
        "json": "application/json",
        "xml": "application/xml",
        "html": "text/html",
        "binary": "application/x-msgpack",
    }

    @staticmethod
    def default_get_headers(binary=False, version=None):
        headers = HttpUtils.default_headers(version=version)
        if binary:
            headers["Accept"] = HttpUtils.mime_types['binary']
        else:
            headers["Accept"] = HttpUtils.mime_types['json']
        return headers

    @staticmethod
    def default_post_headers(binary=False, version=None):
        headers = HttpUtils.default_get_headers(binary=binary, version=version)
        headers["Content-Type"] = headers["Accept"]
        return headers

    @staticmethod
    def get_host_header(host):
        return {
            'Host': host,
        }

    @staticmethod
    def default_headers(version=None):
        if version is None:
            version = ably.api_version
        return {
            "X-Ably-Version": version,
            "Ably-Agent": 'ably-python/%s python/%s' % (ably.lib_version, platform.python_version())
        }

    @staticmethod
    def get_query_params(options):
        params = {}

        if options.add_request_ids:
            params['request_id'] = base64.urlsafe_b64encode(os.urandom(12)).decode('ascii')

        return params



from __future__ import annotations


class ChannelDetails:

    def __init__(self, channel_id, status):
        self.__channel_id = channel_id
        self.__status = status

    @property
    def channel_id(self) -> str:
        return self.__channel_id

    @property
    def status(self) -> ChannelStatus:
        return self.__status

    @staticmethod
    def from_dict(obj):
        kwargs = {
            'channel_id': obj.get("channelId"),
            'status': ChannelStatus.from_dict(obj.get("status"))
        }

        return ChannelDetails(**kwargs)


class ChannelStatus:

    def __init__(self, is_active, occupancy):
        self.__is_active = is_active
        self.__occupancy = occupancy

    @property
    def is_active(self) -> bool:
        return self.__is_active

    @property
    def occupancy(self) -> ChannelOccupancy:
        return self.__occupancy

    @staticmethod
    def from_dict(obj):
        kwargs = {
            'is_active': obj.get("isActive"),
            'occupancy': ChannelOccupancy.from_dict(obj.get("occupancy"))
        }

        return ChannelStatus(**kwargs)


class ChannelOccupancy:

    def __init__(self, metrics):
        self.__metrics = metrics

    @property
    def metrics(self) -> ChannelMetrics:
        return self.__metrics

    @staticmethod
    def from_dict(obj):
        kwargs = {
            'metrics': ChannelMetrics.from_dict(obj.get("metrics"))
        }

        return ChannelOccupancy(**kwargs)


class ChannelMetrics:

    def __init__(self, connections, presence_connections, presence_members,
                 presence_subscribers, publishers, subscribers):
        self.__connections = connections
        self.__presence_connections = presence_connections
        self.__presence_members = presence_members
        self.__presence_subscribers = presence_subscribers
        self.__publishers = publishers
        self.__subscribers = subscribers

    @property
    def connections(self) -> int:
        return self.__connections

    @property
    def presence_connections(self) -> int:
        return self.__presence_connections

    @property
    def presence_members(self) -> int:
        return self.__presence_members

    @property
    def presence_subscribers(self) -> int:
        return self.__presence_subscribers

    @property
    def publishers(self) -> int:
        return self.__publishers

    @property
    def subscribers(self) -> int:
        return self.__subscribers

    @staticmethod
    def from_dict(obj):
        kwargs = {
            'connections': obj.get("connections"),
            'presence_connections': obj.get("presenceConnections"),
            'presence_members': obj.get("presenceMembers"),
            'presence_subscribers': obj.get("presenceSubscribers"),
            'publishers': obj.get("publishers"),
            'subscribers': obj.get("subscribers")
        }

        return ChannelMetrics(**kwargs)



from ably.util import case


class PushChannelSubscription:

    def __init__(self, channel, device_id=None, client_id=None, app_id=None):
        if not device_id and not client_id:
            raise ValueError('missing expected device or client id')

        if device_id and client_id:
            raise ValueError('both device and client id given, only one expected')

        self.__channel = channel
        self.__device_id = device_id
        self.__client_id = client_id
        self.__app_id = app_id

    @property
    def channel(self):
        return self.__channel

    @property
    def device_id(self):
        return self.__device_id

    @property
    def client_id(self):
        return self.__client_id

    @property
    def app_id(self):
        return self.__app_id

    def as_dict(self):
        keys = ['channel', 'device_id', 'client_id', 'app_id']

        obj = {}
        for key in keys:
            value = getattr(self, key)
            if value is not None:
                key = case.snake_to_camel(key)
                obj[key] = value

        return obj

    @classmethod
    def from_dict(cls, obj):
        obj = {case.camel_to_snake(key): value for key, value in obj.items()}
        return cls(**obj)

    @classmethod
    def from_array(cls, array):
        return [cls.from_dict(d) for d in array]

    @classmethod
    def factory(cls, subscription):
        if isinstance(subscription, cls):
            return subscription

        return cls.from_dict(subscription)


def channel_subscriptions_response_processor(response):
    native = response.to_native()
    return PushChannelSubscription.from_array(native)


def channels_response_processor(response):
    native = response.to_native()
    return native



from dataclasses import dataclass


@dataclass()
class ConnectionDetails:
    connection_state_ttl: int
    max_idle_interval: int
    connection_key: str

    def __init__(self, connection_state_ttl: int, max_idle_interval: int,
                 connection_key: str, client_id: str):
        self.connection_state_ttl = connection_state_ttl
        self.max_idle_interval = max_idle_interval
        self.connection_key = connection_key
        self.client_id = client_id

    @staticmethod
    def from_dict(json_dict: dict):
        return ConnectionDetails(json_dict.get('connectionStateTtl'), json_dict.get('maxIdleInterval'),
                                 json_dict.get('connectionKey'), json_dict.get('clientId'))



from ably.sync.util import case


DevicePushTransportType = {'fcm', 'gcm', 'apns', 'web'}
DevicePlatform = {'android', 'ios', 'browser'}
DeviceFormFactor = {'phone', 'tablet', 'desktop', 'tv', 'watch', 'car', 'embedded', 'other'}


class DeviceDetails:

    def __init__(self, id, client_id=None, form_factor=None, metadata=None,
                 platform=None, push=None, update_token=None, app_id=None,
                 device_identity_token=None, modified=None, device_secret=None):

        if push:
            recipient = push.get('recipient')
            if recipient:
                transport_type = recipient.get('transportType')
                if transport_type is not None and transport_type not in DevicePushTransportType:
                    raise ValueError('unexpected transport type {}'.format(transport_type))

        if platform is not None and platform not in DevicePlatform:
            raise ValueError('unexpected platform {}'.format(platform))

        if form_factor is not None and form_factor not in DeviceFormFactor:
            raise ValueError('unexpected form factor {}'.format(form_factor))

        self.__id = id
        self.__client_id = client_id
        self.__form_factor = form_factor
        self.__metadata = metadata
        self.__platform = platform
        self.__push = push
        self.__update_token = update_token
        self.__app_id = app_id
        self.__device_identity_token = device_identity_token
        self.__modified = modified
        self.__device_secret = device_secret

    @property
    def id(self):
        return self.__id

    @property
    def client_id(self):
        return self.__client_id

    @property
    def form_factor(self):
        return self.__form_factor

    @property
    def metadata(self):
        return self.__metadata

    @property
    def platform(self):
        return self.__platform

    @property
    def push(self):
        return self.__push

    @property
    def update_token(self):
        return self.__update_token

    @property
    def app_id(self):
        return self.__app_id

    @property
    def device_identity_token(self):
        return self.__device_identity_token

    @property
    def modified(self):
        return self.__modified

    @property
    def device_secret(self):
        return self.__device_secret

    def as_dict(self):
        keys = ['id', 'client_id', 'form_factor', 'metadata', 'platform',
                'push', 'update_token', 'app_id', 'device_identity_token', 'modified', 'device_secret']

        obj = {}
        for key in keys:
            value = getattr(self, key)
            if value is not None:
                key = case.snake_to_camel(key)
                obj[key] = value

        return obj

    @classmethod
    def from_dict(cls, obj):
        obj = {case.camel_to_snake(key): value for key, value in obj.items()}
        return cls(**obj)

    @classmethod
    def from_array(cls, array):
        return [cls.from_dict(d) for d in array]

    @classmethod
    def factory(cls, device):
        if isinstance(device, cls):
            return device

        return cls.from_dict(device)


def device_details_response_processor(response):
    native = response.to_native()
    return DeviceDetails.from_array(native)



import calendar
import logging
from urllib.parse import urlencode

from ably.http.http import Request
from ably.util import case

log = logging.getLogger(__name__)


def format_time_param(t):
    try:
        return '%d' % (calendar.timegm(t.utctimetuple()) * 1000)
    except Exception:
        return str(t)


def format_params(params=None, direction=None, start=None, end=None, limit=None, **kw):
    if params is None:
        params = {}

    for key, value in kw.items():
        if value is not None:
            key = case.snake_to_camel(key)
            params[key] = value

    if direction:
        params['direction'] = str(direction)
    if start:
        params['start'] = format_time_param(start)
    if end:
        params['end'] = format_time_param(end)
    if limit:
        if limit > 1000:
            raise ValueError("The maximum allowed limit is 1000")
        params['limit'] = '%d' % limit

    if 'start' in params and 'end' in params and params['start'] > params['end']:
        raise ValueError("'end' parameter has to be greater than or equal to 'start'")

    return '?' + urlencode(params) if params else ''


class PaginatedResult:
    def __init__(self, http, items, content_type, rel_first, rel_next,
                 response_processor, response):
        self.__http = http
        self.__items = items
        self.__content_type = content_type
        self.__rel_first = rel_first
        self.__rel_next = rel_next
        self.__response_processor = response_processor
        self.response = response

    @property
    def items(self):
        return self.__items

    def has_first(self):
        return self.__rel_first is not None

    def has_next(self):
        return self.__rel_next is not None

    def is_last(self):
        return not self.has_next()

    async def first(self):
        return await self.__get_rel(self.__rel_first) if self.__rel_first else None

    async def next(self):
        return await self.__get_rel(self.__rel_next) if self.__rel_next else None

    async def __get_rel(self, rel_req):
        if rel_req is None:
            return None
        return await self.paginated_query_with_request(self.__http, rel_req, self.__response_processor)

    @classmethod
    async def paginated_query(cls, http, method='GET', url='/', version=None, body=None,
                              headers=None, response_processor=None,
                              raise_on_error=True):
        headers = headers or {}
        req = Request(method, url, version=version, body=body, headers=headers, skip_auth=False,
                      raise_on_error=raise_on_error)
        return await cls.paginated_query_with_request(http, req, response_processor)

    @classmethod
    async def paginated_query_with_request(cls, http, request, response_processor,
                                           raise_on_error=True):
        response = await http.make_request(
            request.method, request.url, version=request.version,
            headers=request.headers, body=request.body,
            skip_auth=request.skip_auth, raise_on_error=request.raise_on_error)

        items = response_processor(response)

        content_type = response.headers['Content-Type']
        links = response.links
        if 'first' in links:
            first_rel_request = request.with_relative_url(links['first']['url'])
        else:
            first_rel_request = None

        if 'next' in links:
            next_rel_request = request.with_relative_url(links['next']['url'])
        else:
            next_rel_request = None

        return cls(http, items, content_type, first_rel_request,
                   next_rel_request, response_processor, response)


class HttpPaginatedResponse(PaginatedResult):
    @property
    def status_code(self):
        return self.response.status_code

    @property
    def success(self):
        status_code = self.status_code
        return 200 <= status_code < 300

    @property
    def error_code(self):
        return self.response.headers.get('X-Ably-Errorcode')

    @property
    def error_message(self):
        return self.response.headers.get('X-Ably-Errormessage')

    @property
    def headers(self):
        return list(self.response.headers.items())



import logging
from typing import Optional
from urllib.parse import urlencode

from ably.http.http import Http
from ably.http.paginatedresult import PaginatedResult, HttpPaginatedResponse
from ably.http.paginatedresult import format_params
from ably.rest.auth import Auth
from ably.rest.channel import Channels
from ably.rest.push import Push
from ably.util.exceptions import AblyException, catch_all
from ably.types.options import Options
from ably.types.stats import stats_response_processor
from ably.types.tokendetails import TokenDetails

log = logging.getLogger(__name__)


class AblyRest:
    """Ably Rest Client"""

    def __init__(self, key: Optional[str] = None, token: Optional[str] = None,
                 token_details: Optional[TokenDetails] = None, **kwargs):
        """Create an AblyRest instance.

        :Parameters:
          **Credentials**
          - `key`: a valid key string

          **Or**
          - `token`: a valid token string
          - `token_details`: an instance of TokenDetails class

          **Optional Parameters**
          - `client_id`: Undocumented
          - `rest_host`: The host to connect to. Defaults to rest.ably.io
          - `environment`: The environment to use. Defaults to 'production'
          - `port`: The port to connect to. Defaults to 80
          - `tls_port`: The tls_port to connect to. Defaults to 443
          - `tls`: Specifies whether the client should use TLS. Defaults
            to True
          - `auth_token`: Undocumented
          - `auth_callback`: Undocumented
          - `auth_url`: Undocumented
          - `keep_alive`: use persistent connections. Defaults to True
        """
        if key is not None and ('key_name' in kwargs or 'key_secret' in kwargs):
            raise ValueError("key and key_name or key_secret are mutually exclusive. "
                             "Provider either a key or key_name & key_secret")
        if key is not None:
            options = Options(key=key, **kwargs)
        elif token is not None:
            options = Options(auth_token=token, **kwargs)
        elif token_details is not None:
            if not isinstance(token_details, TokenDetails):
                raise ValueError("token_details must be an instance of TokenDetails")
            options = Options(token_details=token_details, **kwargs)
        elif not ('auth_callback' in kwargs or 'auth_url' in kwargs or
                  # and don't have both key_name and key_secret
                  ('key_name' in kwargs and 'key_secret' in kwargs)):
            raise ValueError("key is missing. Either an API key, token, or token auth method must be provided")
        else:
            options = Options(**kwargs)

        try:
            self._is_realtime
        except AttributeError:
            self._is_realtime = False

        self.__http = Http(self, options)
        self.__auth = Auth(self, options)
        self.__http.auth = self.__auth

        self.__channels = Channels(self)
        self.__options = options
        self.__push = Push(self)

    async def __aenter__(self):
        return self

    @catch_all
    async def stats(self, direction: Optional[str] = None, start=None, end=None, params: Optional[dict] = None,
                    limit: Optional[int] = None, paginated=None, unit=None, timeout=None):
        """Returns the stats for this application"""
        formatted_params = format_params(params, direction=direction, start=start, end=end, limit=limit, unit=unit)
        url = '/stats' + formatted_params
        return await PaginatedResult.paginated_query(
            self.http, url=url, response_processor=stats_response_processor)

    @catch_all
    async def time(self, timeout: Optional[float] = None) -> float:
        """Returns the current server time in ms since the unix epoch"""
        r = await self.http.get('/time', skip_auth=True, timeout=timeout)
        AblyException.raise_for_response(r)
        return r.to_native()[0]

    @property
    def client_id(self) -> Optional[str]:
        return self.options.client_id

    @property
    def channels(self):
        """Returns the channels container object"""
        return self.__channels

    @property
    def auth(self):
        return self.__auth

    @property
    def http(self):
        return self.__http

    @property
    def options(self):
        return self.__options

    @property
    def push(self):
        return self.__push

    async def request(self, method: str, path: str, version: str, params:
                      Optional[dict] = None, body=None, headers=None):
        if version is None:
            raise AblyException("No version parameter", 400, 40000)

        url = path
        if params:
            url += '?' + urlencode(params)

        def response_processor(response):
            items = response.to_native()
            if not items:
                return []
            if type(items) is not list:
                items = [items]
            return items

        return await HttpPaginatedResponse.paginated_query(
            self.http, method, url, version=version, body=body, headers=headers,
            response_processor=response_processor,
            raise_on_error=False)

    async def __aexit__(self, *excinfo):
        await self.close()

    async def close(self):
        await self.http.close()



import re


first_cap_re = re.compile('(.)([A-Z][a-z]+)')
all_cap_re = re.compile('([a-z0-9])([A-Z])')


def camel_to_snake(name):
    s1 = first_cap_re.sub(r'\1_\2', name)
    return all_cap_re.sub(r'\1_\2', s1).lower()


def snake_to_camel(name):
    name = name.split('_')
    for i in range(1, len(name)):
        name[i] = name[i].title()

    return ''.join(name)



import functools
import logging
import time
import json
from urllib.parse import urljoin

import httpx
import msgpack

from ably.rest.auth import Auth
from ably.http.httputils import HttpUtils
from ably.transport.defaults import Defaults
from ably.util.exceptions import AblyException
from ably.util.helper import is_token_error

log = logging.getLogger(__name__)


def reauth_if_expired(func):
    @functools.wraps(func)
    async def wrapper(rest, *args, **kwargs):
        if kwargs.get("skip_auth"):
            return await func(rest, *args, **kwargs)

        # RSA4b1 Detect expired token to avoid round-trip request
        auth = rest.auth
        token_details = auth.token_details
        if token_details and auth.time_offset is not None and auth.token_details_has_expired():
            await auth.authorize()
            retried = True
        else:
            retried = False

        try:
            return await func(rest, *args, **kwargs)
        except AblyException as e:
            if is_token_error(e) and not retried:
                await auth.authorize()
                return await func(rest, *args, **kwargs)

            raise e

    return wrapper


class Request:
    def __init__(self, method='GET', url='/', version=None, headers=None, body=None,
                 skip_auth=False, raise_on_error=True):
        self.__method = method
        self.__headers = headers or {}
        self.__body = body
        self.__skip_auth = skip_auth
        self.__url = url
        self.__version = version
        self.raise_on_error = raise_on_error

    def with_relative_url(self, relative_url):
        url = urljoin(self.url, relative_url)
        return Request(self.method, url, self.version, self.headers, self.body,
                       self.skip_auth, self.raise_on_error)

    @property
    def method(self):
        return self.__method

    @property
    def url(self):
        return self.__url

    @property
    def headers(self):
        return self.__headers

    @property
    def body(self):
        return self.__body

    @property
    def skip_auth(self):
        return self.__skip_auth

    @property
    def version(self):
        return self.__version


class Response:
    """
    Composition for httpx.Response with delegation
    """

    def __init__(self, response):
        self.__response = response

    def to_native(self):
        content = self.__response.content
        if not content:
            return None

        content_type = self.__response.headers.get('content-type')
        if isinstance(content_type, str):
            if content_type.startswith('application/x-msgpack'):
                return msgpack.unpackb(content)
            elif content_type.startswith('application/json'):
                return self.__response.json()

        raise ValueError("Unsupported content type")

    @property
    def response(self):
        return self.__response

    def __getattr__(self, attr):
        return getattr(self.__response, attr)


class Http:
    CONNECTION_RETRY_DEFAULTS = {
        'http_open_timeout': 4,
        'http_request_timeout': 10,
        'http_max_retry_duration': 15,
    }

    def __init__(self, ably, options):
        options = options or {}
        self.__ably = ably
        self.__options = options
        self.__auth = None
        # Cached fallback host (RSC15f)
        self.__host = None
        self.__host_expires = None
        self.__client = httpx.AsyncClient(http2=True)

    async def close(self):
        await self.__client.aclose()

    def dump_body(self, body):
        if self.options.use_binary_protocol:
            return msgpack.packb(body, use_bin_type=False)
        else:
            return json.dumps(body, separators=(',', ':'))

    def get_rest_hosts(self):
        hosts = self.options.get_rest_hosts()
        host = self.__host or self.options.fallback_realtime_host
        if host is None:
            return hosts

        if time.time() > self.__host_expires:
            self.__host = None
            self.__host_expires = None
            return hosts

        hosts = list(hosts)
        hosts.remove(host)
        hosts.insert(0, host)
        return hosts

    @reauth_if_expired
    async def make_request(self, method, path, version=None, headers=None, body=None,
                           skip_auth=False, timeout=None, raise_on_error=True):

        if body is not None and type(body) not in (bytes, str):
            body = self.dump_body(body)

        if body:
            all_headers = HttpUtils.default_post_headers(self.options.use_binary_protocol, version=version)
        else:
            all_headers = HttpUtils.default_get_headers(self.options.use_binary_protocol, version=version)

        params = HttpUtils.get_query_params(self.options)

        if not skip_auth:
            if self.auth.auth_mechanism == Auth.Method.BASIC and self.preferred_scheme.lower() == 'http':
                raise AblyException(
                    "Cannot use Basic Auth over non-TLS connections",
                    401,
                    40103)
            auth_headers = await self.auth._get_auth_headers()
            all_headers.update(auth_headers)
        if headers:
            all_headers.update(headers)

        timeout = (self.http_open_timeout, self.http_request_timeout)
        http_max_retry_duration = self.http_max_retry_duration
        requested_at = time.time()

        hosts = self.get_rest_hosts()
        for retry_count, host in enumerate(hosts):
            base_url = "%s://%s:%d" % (self.preferred_scheme,
                                       host,
                                       self.preferred_port)
            url = urljoin(base_url, path)

            request = self.__client.build_request(
                method=method,
                url=url,
                content=body,
                params=params,
                headers=all_headers,
                timeout=timeout,
            )
            try:
                response = await self.__client.send(request)
            except Exception as e:
                # if last try or cumulative timeout is done, throw exception up
                time_passed = time.time() - requested_at
                if retry_count == len(hosts) - 1 or time_passed > http_max_retry_duration:
                    raise e
            else:
                try:
                    if raise_on_error:
                        AblyException.raise_for_response(response)

                    # Keep fallback host for later (RSC15f)
                    if retry_count > 0 and host != self.options.get_rest_host():
                        self.__host = host
                        self.__host_expires = time.time() + (self.options.fallback_retry_timeout / 1000.0)

                    return Response(response)
                except AblyException as e:
                    if not e.is_server_error:
                        raise e

                    # if last try or cumulative timeout is done, throw exception up
                    time_passed = time.time() - requested_at
                    if retry_count == len(hosts) - 1 or time_passed > http_max_retry_duration:
                        raise e

    async def delete(self, url, headers=None, skip_auth=False, timeout=None):
        result = await self.make_request('DELETE', url, headers=headers,
                                         skip_auth=skip_auth, timeout=timeout)
        return result

    async def get(self, url, headers=None, skip_auth=False, timeout=None):
        result = await self.make_request('GET', url, headers=headers,
                                         skip_auth=skip_auth, timeout=timeout)
        return result

    async def patch(self, url, headers=None, body=None, skip_auth=False, timeout=None):
        result = await self.make_request('PATCH', url, headers=headers, body=body,
                                         skip_auth=skip_auth, timeout=timeout)
        return result

    async def post(self, url, headers=None, body=None, skip_auth=False, timeout=None):
        result = await self.make_request('POST', url, headers=headers, body=body,
                                         skip_auth=skip_auth, timeout=timeout)
        return result

    async def put(self, url, headers=None, body=None, skip_auth=False, timeout=None):
        result = await self.make_request('PUT', url, headers=headers, body=body,
                                         skip_auth=skip_auth, timeout=timeout)
        return result

    @property
    def auth(self):
        return self.__auth

    @auth.setter
    def auth(self, value):
        self.__auth = value

    @property
    def options(self):
        return self.__options

    @property
    def preferred_host(self):
        return self.options.get_rest_host()

    @property
    def preferred_port(self):
        return Defaults.get_port(self.options)

    @property
    def preferred_scheme(self):
        return Defaults.get_scheme(self.options)

    @property
    def http_open_timeout(self):
        if self.options.http_open_timeout is not None:
            return self.options.http_open_timeout
        return self.CONNECTION_RETRY_DEFAULTS['http_open_timeout']

    @property
    def http_request_timeout(self):
        if self.options.http_request_timeout is not None:
            return self.options.http_request_timeout
        return self.CONNECTION_RETRY_DEFAULTS['http_request_timeout']

    @property
    def http_max_retry_count(self):
        if self.options.http_max_retry_count is not None:
            return self.options.http_max_retry_count
        return self.CONNECTION_RETRY_DEFAULTS['http_max_retry_count']

    @property
    def http_max_retry_duration(self):
        if self.options.http_max_retry_duration is not None:
            return self.options.http_max_retry_duration
        return self.CONNECTION_RETRY_DEFAULTS['http_max_retry_duration']



from collections.abc import MutableMapping
import json
import logging


log = logging.getLogger(__name__)


class Capability(MutableMapping):
    def __init__(self, obj=None):
        if obj is None:
            obj = {}
        self.__dict = dict(obj)
        for k, v in obj.items():
            self[k] = v

    def __eq__(self, other):
        if isinstance(other, Capability):
            return Capability.c14n(self) == Capability.c14n(other)
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, Capability):
            return Capability.c14n(self) != Capability.c14n(other)
        return NotImplemented

    def __getitem__(self, key):
        return self.__dict[key]

    def __iter__(self):
        return iter(self.__dict)

    def __len__(self):
        return len(self.__dict)

    def __contains__(self, key):
        return key in self.__dict

    def __setitem__(self, key, value):
        # validate that the value is a list of ops and that the key is a string
        if not isinstance(key, str):
            raise ValueError('Capability keys must be strings')

        if isinstance(value, str):
            value = [value]

        operations = set()
        for val in iter(value):
            if not isinstance(val, str):
                raise ValueError('Operations must be strings')
            operations.add(val)

        self.__dict[key] = operations

    def __delitem__(self, key):
        del self.__dict[key]

    def setdefault(self, key, default):
        if key not in self:
            self[key] = default
        return self[key]

    def add_resource(self, resource, operations=None):
        if operations is None:
            operations = []
        if isinstance(operations, str):
            operations = [operations]
        self[resource] = list(operations)

    def add_operation_to_resource(self, operation, resource):
        self.setdefault(resource, []).append(operation)

    def __str__(self):
        return Capability.c14n(self)

    def to_dict(self):
        return {k: sorted(v) for k, v in self.items()}

    @staticmethod
    def c14n(capability):
        sorted_ops = capability.to_dict()
        return json.dumps(sorted_ops, sort_keys=True)



from ably.sync.types.connectionstate import ConnectionState
from ably.sync.util.exceptions import AblyException

ConnectionErrors = {
    ConnectionState.DISCONNECTED: AblyException(
        'Connection to server temporarily unavailable',
        400,
        80003,
    ),
    ConnectionState.SUSPENDED: AblyException(
        'Connection to server unavailable',
        400,
        80002,
    ),
    ConnectionState.FAILED: AblyException(
        'Connection failed or disconnected by server',
        400,
        80000,
    ),
    ConnectionState.CLOSING: AblyException(
        'Connection closing',
        400,
        80017,
    ),
    ConnectionState.CLOSED: AblyException(
        'Connection closed',
        400,
        80017,
    ),
}
